@@ -15,14 +15,22 @@ import subprocess
 import functools
 import re
 
-import apt
+import apt_pkg
 
 from UbuntuDrivers import kerneldetection
 
-system_architecture = apt.apt_pkg.get_architectures()[0]
+system_architecture = ''
 lookup_cache = {}
 
 
+def get_apt_arch():
+    '''Cache system architecture'''
+    global system_architecture
+    if not system_architecture:
+        system_architecture = apt_pkg.get_architectures()[0]
+    return system_architecture
+
+
 def system_modaliases(sys_path=None):
     '''Get modaliases present in the system.
 
@@ -70,35 +78,64 @@ def system_modaliases(sys_path=None):
     return aliases
 
 
-def _check_video_abi_compat(apt_cache, record):
+def _check_video_abi_compat(apt_cache, package):
     xorg_video_abi = None
 
+    if package.name.startswith('nvidia-driver-'):
+        xorg_driver_name = package.name.replace('nvidia-driver-', 'xserver-xorg-video-nvidia-')
+        try:
+            package = apt_cache[xorg_driver_name]
+        except KeyError:
+            logging.debug('Cannot find %s package in the cache. Cannot check ABI' % (xorg_driver_name))
+            return True
+
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(package)
+
+    needs_video_abi = False
+    try:
+        for dep_list in candidate.depends_list_str.get('Depends'):
+            for dep_name, dep_ver, dep_op in dep_list:
+                if dep_name.startswith('xorg-video-abi-'):
+                    needs_video_abi = True
+                    break
+    except (KeyError, TypeError):
+        logging.debug('The %s package seems to have no dependencies. Skipping ABI check' % (package))
+        needs_video_abi = False
+
+    if not needs_video_abi:
+        logging.debug('Skipping check for %s since it does not depend on video abi' % package.name)
+        return True
+
     # determine current X.org video driver ABI
     try:
-        for p in apt_cache['xserver-xorg-core'].candidate.provides:
-            if p.startswith('xorg-video-abi-'):
-                xorg_video_abi = p
-                # logging.debug('_check_video_abi_compat(): Current X.org video abi: %s', xorg_video_abi)
-                break
-    except (AttributeError, KeyError):
-        logging.debug('_check_video_abi_compat(): xserver-xorg-core not available, cannot check ABI')
+        xorg_core = apt_cache['xserver-xorg-core']
+    except KeyError:
+        logging.debug('xserver-xorg-core not available, cannot check ABI')
         return True
-    if not xorg_video_abi:
+
+    candidate = depcache.get_candidate_ver(xorg_core)
+
+    for provides_name, provides_ver, p_version in candidate.provides_list:
+        if provides_name.startswith('xorg-video-abi-'):
+            xorg_video_abi = provides_name
+
+    if xorg_video_abi:
+        abi_pkg = apt_cache[xorg_video_abi]
+        for dep in abi_pkg.rev_depends_list:
+            if dep.parent_pkg.name == package.name:
+                return True
+
+        logging.debug('Driver package %s is incompatible with current X.org server ABI %s',
+                      package.name, xorg_video_abi)
+        logging.debug('%s is not in %s' % (package, [x.parent_pkg for x in abi_pkg.rev_depends_list]))
         return False
 
-    try:
-        deps = record['Depends']
-    except KeyError:
-        return True
-    if 'xorg-video-abi-' in deps and xorg_video_abi not in deps:
-        logging.debug('Driver package %s is incompatible with current X.org server ABI %s',
-                      record['Package'], xorg_video_abi)
-        return False
     return True
 
 
 def _apt_cache_modalias_map(apt_cache):
-    '''Build a modalias map from an apt.Cache object.
+    '''Build a modalias map from an apt_pkg.Cache object.
 
     This filters out uninstallable video drivers (i. e. which depend on a video
     ABI that xserver-xorg-core does not provide).
@@ -106,22 +143,27 @@ def _apt_cache_modalias_map(apt_cache):
     Return a map bus -> modalias -> [package, ...], where "bus" is the prefix of
     the modalias up to the first ':' (e. g. "pci" or "usb").
     '''
+    depcache = apt_pkg.DepCache(apt_cache)
+    records = apt_pkg.PackageRecords(apt_cache)
+
     result = {}
-    for package in apt_cache:
+    for package in apt_cache.packages:
         # skip packages without a modalias field
         try:
-            m = package.candidate.record['Modaliases']
+            candidate = depcache.get_candidate_ver(package)
+            records.lookup(candidate.file_list[0])
+            m = records['Modaliases']
+            if not m:
+                continue
         except (KeyError, AttributeError, UnicodeDecodeError):
             continue
 
         # skip foreign architectures, we usually only want native
         # driver packages
-        if (not package.candidate or
-                package.candidate.architecture not in ('all', system_architecture)):
+        if (package.architecture not in ('all', get_apt_arch())):
             continue
 
-        # skip incompatible video drivers
-        if not _check_video_abi_compat(apt_cache, package.candidate.record):
+        if not _check_video_abi_compat(apt_cache, package):
             continue
 
         try:
@@ -138,7 +180,14 @@ def _apt_cache_modalias_map(apt_cache):
             logging.error('Package %s has invalid modalias header: %s' % (
                 package.name, m))
 
-    return result
+    result2 = {}
+    for bus, alias_map in result.items():
+        pat = re.compile(
+            '|'.join([fnmatch.translate(pat) for pat in alias_map.keys()]),
+            re.IGNORECASE)
+        result2[bus] = (pat, alias_map)
+
+    return result2
 
 
 def packages_for_modalias(apt_cache, modalias):
@@ -148,16 +197,18 @@ def packages_for_modalias(apt_cache, modalias):
     '''
     pkgs = set()
 
-    apt_cache_hash = hash(apt_cache)
+    apt_cache_hash = hash(package.get_fullname() for package in apt_cache.packages)
     try:
         cache_map = packages_for_modalias.cache_maps[apt_cache_hash]
     except KeyError:
         cache_map = _apt_cache_modalias_map(apt_cache)
         packages_for_modalias.cache_maps[apt_cache_hash] = cache_map
 
-    bus_map = cache_map.get(modalias.split(':', 1)[0], {})
+    pat, bus_map = cache_map.get(modalias.split(':', 1)[0], (None, {}))
+    if pat is None or not pat.match(modalias):
+        return []
     for alias in bus_map:
-        if fnmatch.fnmatch(modalias.lower(), alias.lower()):
+        if fnmatch.fnmatchcase(modalias.lower(), alias.lower()):
             for p in bus_map[alias]:
                 pkgs.add(p)
 
@@ -167,8 +218,10 @@ def packages_for_modalias(apt_cache, modalias):
 packages_for_modalias.cache_maps = {}
 
 
-def _is_package_free(pkg):
-    assert pkg.candidate is not None
+def _is_package_free(apt_cache, pkg):
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(pkg)
+    assert candidate is not None
     # it would be better to check the actual license, as we do not have
     # the component for third-party packages; but this is the best we can do
     # at the moment
@@ -176,52 +229,75 @@ def _is_package_free(pkg):
     # We can assume the NVIDIA packages to be non-free
     if pkg.name.startswith('nvidia'):
         return False
-    for o in pkg.candidate.origins:
-        if o.component in ('restricted', 'multiverse'):
-            return False
+
+    records = apt_pkg.PackageRecords(apt_cache)
+    records.lookup(candidate.file_list[0])
+
+    for pfile, _ in pkg.version_list[0].file_list:
+        if not pfile.component:
+            # This is probably from the test suite
+            try:
+                component = records['Component']
+                return (component not in ('restricted', 'multiverse'))
+            except KeyError:
+                return False
+        else:
+            if pfile.component in ('restricted', 'multiverse'):
+                return False
     return True
 
 
-def _is_package_from_distro(pkg):
-    if pkg.candidate is None:
+def _is_package_from_distro(apt_cache, pkg):
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(pkg)
+    if candidate is None:
         return False
 
-    for o in pkg.candidate.origins:
-        if o.origin == 'Ubuntu':
-            return True
-    return False
+    try:
+        return (candidate.file_list[0][0].origin == 'Ubuntu')
+    except KeyError:
+        return False
 
 
-def _pkg_get_module(pkg):
+def _pkg_get_module(apt_cache, pkg):
     '''Determine module name from apt Package object'''
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(pkg)
+    records = apt_pkg.PackageRecords(apt_cache)
+    records.lookup(candidate.file_list[0])
 
     try:
-        m = pkg.candidate.record['Modaliases']
+        m = records['Modaliases']
     except (KeyError, AttributeError):
         logging.debug('_pkg_get_module %s: package has no Modaliases header, cannot determine module', pkg.name)
         return None
 
     paren = m.find('(')
     if paren <= 0:
-        logging.warning('_pkg_get_module %s: package has invalid Modaliases header, cannot determine module', pkg.name)
+        logging.debug('_pkg_get_module %s: package has invalid Modaliases header, cannot determine module', pkg.name)
         return None
 
     module = m[:paren]
     return module
 
 
-def _pkg_get_support(pkg):
+def _pkg_get_support(apt_cache, pkg):
     '''Determine support level from apt Package object'''
 
+    depcache = apt_pkg.DepCache(apt_cache)
+    candidate = depcache.get_candidate_ver(pkg)
+    records = apt_pkg.PackageRecords(apt_cache)
+    records.lookup(candidate.file_list[0])
+
     try:
-        support = pkg.candidate.record['Support']
+        support = records['Support']
     except (KeyError, AttributeError):
         logging.debug('_pkg_get_support %s: package has no Support header, cannot determine support level', pkg.name)
         return None
 
     if support not in ('PB', 'NFB', 'LTSB', 'Legacy'):
-        logging.warning('_pkg_get_support %s: package has invalid Support %s'
-                        'header, cannot determine support level', pkg.name, support)
+        logging.debug('_pkg_get_support %s: package has invalid Support %s'
+                      'header, cannot determine support level', pkg.name, support)
         return None
 
     return support
@@ -247,10 +323,10 @@ def is_wayland_session():
     return os.environ.get('WAYLAND_DISPLAY') is not None
 
 
-def _is_manual_install(pkg):
+def _is_manual_install(apt_cache, pkg):
     '''Determine if the kernel module from an apt.Package is manually installed.'''
 
-    if pkg.installed:
+    if pkg.current_ver:
         return False
 
     # special case, as our packages suffix the kmod with _version
@@ -259,7 +335,7 @@ def _is_manual_install(pkg):
     elif pkg.name.startswith('fglrx'):
         module = 'fglrx'
     else:
-        module = _pkg_get_module(pkg)
+        module = _pkg_get_module(apt_cache, pkg)
 
     if not module:
         return False
@@ -328,7 +404,7 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
     queries apt about which packages provide drivers for those. It also adds
     available packages from detect_plugin_packages().
 
-    If you already have an apt.Cache() object, you should pass it as an
+    If you already have an apt_pkg.Cache() object, you should pass it as an
     argument for efficiency. If not given, this function creates a temporary
     one by itself.
 
@@ -362,7 +438,7 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
 
     if not apt_cache:
         try:
-            apt_cache = apt.Cache()
+            apt_cache = apt_pkg.Cache(None)
         except Exception as ex:
             logging.error(ex)
             return {}
@@ -370,16 +446,16 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
     packages = {}
     for alias, syspath in modaliases.items():
         for p in packages_for_modalias(apt_cache, alias):
-            if freeonly and not _is_package_free(p):
+            if freeonly and not _is_package_free(apt_cache, p):
                 continue
             if not include_oem and fnmatch.fnmatch(p.name, 'oem-*-meta'):
                 continue
             packages[p.name] = {
                     'modalias': alias,
                     'syspath': syspath,
-                    'free': _is_package_free(p),
-                    'from_distro': _is_package_from_distro(p),
-                    'support': _pkg_get_support(p),
+                    'free': _is_package_free(apt_cache, p),
+                    'from_distro': _is_package_from_distro(apt_cache, p),
+                    'support': _pkg_get_support(apt_cache, p),
                     'runtimepm': _is_runtimepm_supported(p, alias)
                 }
             (vendor, model) = _get_db_name(syspath, alias)
@@ -404,12 +480,15 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
     # add available packages which need custom detection code
     for plugin, pkgs in detect_plugin_packages(apt_cache).items():
         for p in pkgs:
-            apt_p = apt_cache[p]
-            packages[p] = {
-                    'free': _is_package_free(apt_p),
-                    'from_distro': _is_package_from_distro(apt_p),
-                    'plugin': plugin,
-                }
+            try:
+                apt_p = apt_cache[p]
+                packages[p] = {
+                        'free': _is_package_free(apt_cache, apt_p),
+                        'from_distro': _is_package_from_distro(apt_cache, apt_p),
+                        'plugin': plugin,
+                    }
+            except KeyError:
+                logging.debug('Package %s plugin not available. Skipping.' % p)
 
     return packages
 
@@ -426,14 +505,15 @@ def _get_vendor_model_from_alias(alias):
 
 
 def _get_headless_no_dkms_metapackage(pkg, apt_cache):
-    assert pkg.candidate is not None
+    assert pkg is not None
     metapackage = None
     '''Return headless-no-dkms metapackage from the main metapackage.
 
     This is useful when dealing with packages such as nvidia-driver-$flavour
     whose headless-no-dkms metapackage would be nvidia-headless-no-dkms-$flavour
     '''
-    name = pkg.shortname
+    depcache = apt_pkg.DepCache(apt_cache)
+    name = pkg.name
 
     pattern = re.compile('nvidia-([0-9]+)')
     match = pattern.match(name)
@@ -451,11 +531,12 @@ def _get_headless_no_dkms_metapackage(pkg, apt_cache):
     candidate = 'nvidia-headless-no-dkms-%s' % (candidate_flavour)
 
     try:
-        package = apt_cache.__getitem__(candidate)
+        package = apt_cache[candidate]
         # skip foreign architectures, we usually only want native
         # driver packages
-        if (package.candidate and
-                package.candidate.architecture in ('all', system_architecture)):
+        package_candidate = depcache.get_candidate_ver(package)
+        if (candidate and
+                package_candidate.arch in ('all', get_apt_arch())):
             metapackage = candidate
     except KeyError:
         pass
@@ -470,7 +551,7 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
     queries apt about which packages provide hardware enablement support for
     those.
 
-    If you already have an apt.Cache() object, you should pass it as an
+    If you already have an apt_pkg.Cache() object, you should pass it as an
     argument for efficiency. If not given, this function creates a temporary
     one by itself.
 
@@ -502,7 +583,7 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
 
     if not apt_cache:
         try:
-            apt_cache = apt.Cache()
+            apt_cache = apt_pkg.Cache(None)
         except Exception as ex:
             logging.error(ex)
             return {}
@@ -515,10 +596,10 @@ def system_device_specific_metapackages(apt_cache=None, sys_path=None, include_o
             packages[p.name] = {
                     'modalias': alias,
                     'syspath': syspath,
-                    'free': _is_package_free(p),
-                    'from_distro': _is_package_from_distro(p),
+                    'free': _is_package_free(apt_cache, p),
+                    'from_distro': _is_package_from_distro(apt_cache, p),
                     'recommended': True,
-                    'support': _pkg_get_support(p),
+                    'support': _pkg_get_support(apt_cache, p),
                 }
 
     return packages
@@ -531,7 +612,7 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
     queries apt about which packages provide drivers for those. Finally, it looks
     for the correct metapackage, by calling _get_headless_no_dkms_metapackage().
 
-    If you already have an apt.Cache() object, you should pass it as an
+    If you already have an apt_pkg.Cache() object, you should pass it as an
     argument for efficiency. If not given, this function creates a temporary
     one by itself.
 
@@ -563,7 +644,7 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
 
     if not apt_cache:
         try:
-            apt_cache = apt.Cache()
+            apt_cache = apt_pkg.Cache(None)
         except Exception as ex:
             logging.error(ex)
             return {}
@@ -577,9 +658,9 @@ def system_gpgpu_driver_packages(apt_cache=None, sys_path=None):
                 packages[p.name] = {
                         'modalias': alias,
                         'syspath': syspath,
-                        'free': _is_package_free(p),
-                        'from_distro': _is_package_from_distro(p),
-                        'support': _pkg_get_support(p),
+                        'free': _is_package_free(apt_cache, p),
+                        'from_distro': _is_package_from_distro(apt_cache, p),
+                        'support': _pkg_get_support(apt_cache, p),
                     }
                 if vendor is not None:
                     packages[p.name]['vendor'] = vendor
@@ -614,7 +695,7 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     adds available packages from detect_plugin_packages(), using the name of
     the detction plugin as device name.
 
-    If you already have an apt.Cache() object, you should pass it as an
+    If you already have an apt_pkg.Cache() object, you should pass it as an
     argument for efficiency. If not given, this function creates a temporary
     one by itself.
 
@@ -662,7 +743,7 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     result = {}
     if not apt_cache:
         try:
-            apt_cache = apt.Cache()
+            apt_cache = apt_pkg.Cache(None)
         except Exception as ex:
             logging.error(ex)
             return {}
@@ -687,7 +768,7 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     # packages are "manually installed"
     for driver, info in result.items():
         for pkg in info['drivers']:
-            if not _is_manual_install(apt_cache[pkg]):
+            if not _is_manual_install(apt_cache, apt_cache[pkg]):
                 break
         else:
             info['manual_install'] = True
@@ -698,6 +779,84 @@ def system_device_drivers(apt_cache=None, sys_path=None, freeonly=False):
     return result
 
 
+def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_oem=True, driver_string=''):
+    '''Return the list of packages that should be installed'''
+    packages = system_driver_packages(
+        apt_cache, sys_path, freeonly=free_only,
+        include_oem=include_oem)
+    packages = auto_install_filter(packages, driver_string)
+    if not packages:
+        logging.debug('No drivers found for installation.')
+        return packages
+
+    depcache = apt_pkg.DepCache(apt_cache)
+    records = apt_pkg.PackageRecords(apt_cache)
+
+    # ignore packages which are already installed
+    to_install = []
+    for p in packages:
+        package_obj = apt_cache[p]
+        if not package_obj.current_ver:
+            to_install.append(p)
+
+            candidate = depcache.get_candidate_ver(package_obj)
+            records.lookup(candidate.file_list[0])
+
+            # See if runtimepm is supported
+            if records['runtimepm']:
+                # Create a file for nvidia-prime
+                try:
+                    pm_fd = open('/run/nvidia_runtimepm_supported', 'w')
+                    pm_fd.write('\n')
+                    pm_fd.close()
+                except PermissionError:
+                    # No need to error out here, since package
+                    # installation will fail
+                    pass
+            # Add the matching linux modules package when available
+            try:
+                modules_package = get_linux_modules_metapackage(apt_cache, p)
+                if modules_package and not apt_cache[modules_package].current_ver:
+                    to_install.append(modules_package)
+            except KeyError:
+                pass
+
+    return to_install
+
+
+def nvidia_desktop_pre_installation_hook(to_install):
+    '''Applies changes that need to happen before installing the NVIDIA drivers'''
+    with_nvidia_kms = False
+
+    # Enable KMS if nvidia >= 470
+    for package_name in to_install:
+        if package_name.startswith('nvidia-driver-'):
+            try:
+                version = int(package_name.split('-')[-1])
+            except ValueError:
+                pass
+            finally:
+                with_nvidia_kms = version >= 470
+
+    if with_nvidia_kms:
+        set_nvidia_kms(1)
+
+
+def nvidia_desktop_post_installation_hook():
+    # If we are dealing with NVIDIA PRIME, and runtimepm
+    # is supported, enable it
+    if os.path.isfile('/run/nvidia_runtimepm_supported'):
+        logging.debug('Trying to select the on-demand PRIME profile')
+        try:
+            subprocess.call(['/usr/bin/prime-select', 'on-demand'])
+        except FileNotFoundError:
+            pass
+
+        # Create the override file for gpu-manager
+        with open('/etc/u-d-c-nvidia-runtimepm-override', 'w') as f:
+            f.write('# File created by ubuntu-drivers\n')
+
+
 class _GpgpuDriver(object):
 
     def __init__(self, vendor=None, flavour=None):
@@ -875,7 +1034,7 @@ def detect_plugin_packages(apt_cache=None):
     returned lists for packages which are available for installation, and
     return the joined results.
 
-    If you already have an existing apt.Cache() object, you can pass it as an
+    If you already have an existing apt_pkg.Cache() object, you can pass it as an
     argument for efficiency.
 
     Return pluginname -> [package, ...] map.
@@ -889,7 +1048,7 @@ def detect_plugin_packages(apt_cache=None):
 
     if apt_cache is None:
         try:
-            apt_cache = apt.Cache()
+            apt_cache = apt_pkg.Cache(None)
         except Exception as ex:
             logging.error(ex)
             return {}
@@ -917,10 +1076,11 @@ def detect_plugin_packages(apt_cache=None):
                 continue
 
             for pkg in result:
-                if pkg in apt_cache and apt_cache[pkg].candidate:
-                    if _check_video_abi_compat(apt_cache, apt_cache[pkg].candidate.record):
+                try:
+                    package = apt_cache[pkg]
+                    if _check_video_abi_compat(apt_cache, package):
                         packages.setdefault(fname, []).append(pkg)
-                else:
+                except KeyError:
                     logging.debug('Ignoring unavailable package %s from plugin %s', pkg, plugin)
 
     return packages
@@ -1076,41 +1236,34 @@ def get_linux(apt_cache):
     return kernel_detection.get_linux_metapackage()
 
 
-def find_reverse_dependencies(apt_cache, package, prefix):
-    '''Return the reverse dependencies for a package'''
-    # prefix to restrict the searching
-    # package we want reverse dependencies for
-    deps = set()
-    for pkg in apt_cache:
-        if pkg.name.startswith(prefix):
-            dependencies = []
-            if pkg.candidate:
-                dependencies.extend(pkg.candidate.dependencies)
-            if pkg.installed:
-                dependencies.extend(pkg.installed.dependencies)
-
-            for ordep in dependencies:
-                for dep in ordep:
-                    if dep.rawtype != 'Depends':
-                        continue
-                    if dep.name == package:
-                        deps.add(pkg.name)
-    return list(deps)
-
-
 def get_linux_image_from_meta(apt_cache, pkg):
-    if apt_cache[pkg].candidate:
-        record = apt_cache[pkg].candidate.record
+    depcache = apt_pkg.DepCache(apt_cache)
 
     try:
-        deps = record['Depends']
+        candidate = depcache.get_candidate_ver(apt_cache[pkg])
     except KeyError:
+        logging.debug('No candidate for %s found' % pkg)
         return None
 
-    deps_list = deps.strip().split(', ')
-    for dep in deps_list:
-        if dep.startswith('linux-image-'):
-            return dep
+    try:
+        for dep_list in candidate.depends_list_str.get('Depends'):
+            for dep_name, dep_ver, dep_op in dep_list:
+                if dep_name.startswith('linux-image-'):
+                    return dep_name
+    except (KeyError, TypeError):
+        logging.debug('Could not check dependencies for %s package' % (pkg))
+    # if apt_cache[pkg].candidate:
+    #     record = apt_cache[pkg].candidate.record
+
+    # try:
+    #     deps = record['Depends']
+    # except KeyError:
+    #     return None
+
+    # deps_list = deps.strip().split(', ')
+    # for dep in deps_list:
+    #     if dep.startswith('linux-image-'):
+    #         return dep
 
     return None
 
@@ -1122,6 +1275,8 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     linux_flavour = ''
     linux_modules_match = ''
 
+    depcache = apt_pkg.DepCache(apt_cache)
+
     if 'nvidia' not in candidate:
         logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping', candidate)
         return metapackage
@@ -1152,20 +1307,22 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     linux_modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_flavour)
 
     try:
-        package = apt_cache.__getitem__(linux_modules_candidate)
+        package = apt_cache[linux_modules_candidate]
         # skip foreign architectures, we usually only want native
-        if (package.candidate and
-                package.candidate.architecture in ('all', system_architecture)):
+        package_candidate = depcache.get_candidate_ver(package)
+
+        if (package_candidate and package_candidate.arch in ('all', get_apt_arch())):
             linux_version = get_linux_version(apt_cache)
             linux_modules_abi_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_version)
             logging.debug('linux_modules_abi_candidate: %s' % (linux_modules_abi_candidate))
 
             # Let's check if there is a candidate that is specific to
             # our kernel ABI. If not, things will fail.
-            abi_specific = apt_cache.__getitem__(linux_modules_abi_candidate)
+            abi_specific = apt_cache[linux_modules_abi_candidate]
             # skip foreign architectures, we usually only want native
-            if (abi_specific.candidate and
-                    abi_specific.candidate.architecture in ('all', system_architecture)):
+            abi_specific_candidate = depcache.get_candidate_ver(abi_specific)
+            if (abi_specific_candidate and
+                    abi_specific_candidate.arch in ('all', get_apt_arch())):
                 logging.debug('Found ABI compatible %s' % (linux_modules_abi_candidate))
                 linux_modules_match = linux_modules_candidate
     except KeyError:
@@ -1177,7 +1334,9 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     if linux_modules_match:
         # Look for the metapackage in the reverse
         # dependencies
-        reverse_deps = find_reverse_dependencies(apt_cache, linux_modules_match, 'linux-modules-nvidia-')
+        reverse_deps = [dep.parent_pkg.name for dep in apt_cache[linux_modules_match].rev_depends_list
+                        if dep.parent_pkg.name.startswith('linux-modules-nvidia-')]
+
         pick = ''
         modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour,
                                                             get_linux_image(apt_cache).replace('linux-image-', ''))
@@ -1194,10 +1353,12 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     logging.debug('Falling back to %s' % (dkms_package))
 
     try:
-        package = apt_cache.__getitem__(dkms_package)
+        package = apt_cache[dkms_package]
+        package_candidate = depcache.get_candidate_ver(package)
+
         # skip foreign architectures, we usually only want native
-        if (package.candidate and
-                package.candidate.architecture in ('all', system_architecture)):
+        if (package_candidate and
+                package_candidate.arch in ('all', get_apt_arch())):
             metapackage = dkms_package
     except KeyError:
         logging.error('No "%s" can be found.', dkms_package)
